/* Copyright (C) 2009 Papavasileiou Dimitris                             
 *                                                                      
 * This program is free software: you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation, either version 3 of the License, or    
 * (at your option) any later version.                                  
 *                                                                      
 * This program is distributed in the hope that it will be useful,      
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        
 * GNU General Public License for more details.                         
 *                                                                      
 * You should have received a copy of the GNU General Public License    
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <lua.h>
#include <lauxlib.h>
#include <GL/gl.h>

#include "lambert.h"
#include "phong.h"
#include "anisotropic.h"
#include "glossy.h"
#include "cook.h"
#include "prototype.h"
#include "minnaert.h"
#include "fresnel.h"
#include "oren.h"
#include "light.h"
#include "fog.h"
#include "ambient.h"
#include "sprites.h"
#include "cloth.h"

static char *decapitalize (char *s)
{
    s[0] = tolower(s[0]);

    return s;
}

static int generic_tostring(lua_State *L)
{
    id object;

    object = *(id *)lua_touserdata(L, 1);
    lua_pushstring(L, [object name]);
   
    return 1;
}

static int generic_index(lua_State *L)
{
    id object;

    object = *(id *)lua_touserdata(L, 1);
    
    [object get];
    
    return 1;
}

static int generic_newindex(lua_State *L)
{
    id object;

    object = *(id *)lua_touserdata(L, 1);
    
    [object set];

    return 0;
}

static int generic_gc(lua_State *L)
{
    id object;

    object = *(id *)lua_touserdata(L, 1);

    [object free];

    return 0;
}

static int constructnode (lua_State *L)
{
    Class class;
    id object;

    lua_pushvalue (L, lua_upvalueindex (1));
    class = (Class)lua_touserdata(L, -1);
    lua_pop(L, 1);

    /* Create the userdata... */
    
    object = [[class alloc] init];
    *(id *)lua_newuserdata(L, sizeof(id)) = object;
    
    lua_newtable (L);
    lua_pushstring(L, "__index");
    lua_pushcfunction(L, (lua_CFunction)generic_index);
    lua_settable(L, -3);
    lua_pushstring(L, "__newindex");
    lua_pushcfunction(L, (lua_CFunction)generic_newindex);
    lua_settable(L, -3);
    lua_pushstring(L, "__tostring");
    lua_pushcfunction(L, (lua_CFunction)generic_tostring);
    lua_settable(L, -3);
    lua_pushstring(L, "__gc");
    lua_pushcfunction(L, (lua_CFunction)generic_gc);
    lua_settable(L, -3);
    lua_setmetatable(L, -2);
    
    /* ...and initialize it. */

    if(lua_istable(L, 1)) {
	lua_pushnil(L);
	
	while(lua_next(L, 1)) {
	    lua_pushvalue(L, -2);
	    lua_insert(L, -2);
	    lua_settable(L, 2);
	}
    }

    /* Add a weak reference. */
    
    lua_pushstring (L, "userdata");
    lua_gettable (L, LUA_REGISTRYINDEX);
    lua_pushlightuserdata (L, object);
    lua_pushvalue (L, 2);
    lua_settable (L, -3);
    lua_pop(L, 1);

    return 1;
}

int luaopen_shading (lua_State *L)
{
    int i;

    Class surfaces[] = {
	[Light class], [Lambert class],
	[Phong class], [Minnaert class],
	[Oren class], [Ambient class],
	[Fog class],
	[Fresnel class], [Anisotropic class],
	[Prototype class], [Glossy class],
	[Cook class], [Sprites class],
	[Cloth class],
    };	
    
    /* And the surface node constructors. */

    lua_newtable (L);
    
    for (i = 0 ; i < sizeof(surfaces) / sizeof(surfaces[0]) ; i += 1) {
	lua_pushlightuserdata (L, surfaces[i]);
	lua_pushcclosure (L, constructnode, 1);
	lua_setfield(L, -2, decapitalize(strdupa([surfaces[i] name])));
    }

    lua_setglobal (L, lua_tostring (L, 1));

    return 0;
}
