/* Copyright (C) 2009 Papavasileiou Dimitris                             
 *                                                                      
 * This program is free software: you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation, either version 3 of the License, or    
 * (at your option) any later version.                                  
 *                                                                      
 * This program is distributed in the hope that it will be useful,      
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        
 * GNU General Public License for more details.                         
 *                                                                      
 * You should have received a copy of the GNU General Public License    
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <ode/ode.h>
#include <lua.h>
#include <lualib.h>
#include <lauxlib.h>

#include "spherical.h"

@implementation Spherical

-(Joint *) init
{
    char *list[] = {
	"anchor", "axes", "motor", "state", "stops", "tolerance"
    };

    self->joint = dJointCreateBall (_WORLD, NULL);
    self->amotor = dJointCreateAMotor (_WORLD, NULL);

    self->anchor[0] = 0;
    self->anchor[1] = 0;
    self->anchor[2] = 0;

    self->axes[0][0] = 1;
    self->axes[0][1] = 0;
    self->axes[0][2] = 0;

    self->axes[2][0] = 0;
    self->axes[2][1] = 0;
    self->axes[2][2] = 1;
    
    self->stops[0][0] = -dInfinity;
    self->stops[0][1] = dInfinity;

    self->stops[1][0] = -dInfinity;
    self->stops[1][1] = dInfinity;

    self->stops[2][0] = -dInfinity;
    self->stops[2][1] = dInfinity;
    
    dJointSetAMotorAxis ([self amotor], 0, 0, 1, 0, 0);
    dJointSetAMotorAxis ([self amotor], 2, 0, 0, 0, 1);

    [super init];
    [self add: sizeof (list) / sizeof (char *) Properties: list];

    return self;
}

-(dJointID) amotor
{
    return self->amotor;
}

-(void) free
{
    dJointDestroy (self->amotor);

    [super free];
}

-(void) update
{
    dBodyID a, b;

    a = dJointGetBody ([self joint], 0);
    b = dJointGetBody ([self joint], 1);

    /* If the joint is already connected update the
       anchor and axes. */
    
    if (a || b) {
	dJointGetBallAnchor ([self joint], self->anchor);
	dJointGetAMotorAxis ([self amotor], 0, self->axes[0]);
	dJointGetAMotorAxis ([self amotor], 2, self->axes[2]);
    }

    [super update];

    a = dJointGetBody ([self joint], 0);
    b = dJointGetBody ([self joint], 1);

    if (a || b) {
    	dJointSetBallAnchor ([self joint],
			     self->anchor[0],
			     self->anchor[1],
			     self->anchor[2]);

    	dJointAttach ([self amotor], a, b);
    	dJointSetAMotorMode (self->amotor, dAMotorEuler);

    	dJointSetAMotorAxis ([self amotor], 0, a ? 1 : 0,
			     self->axes[0][0],
			     self->axes[0][1],
			     self->axes[0][2]);
	dJointSetAMotorAxis ([self amotor], 2, b ? 2 : 0,
			     self->axes[2][0],
			     self->axes[2][1],
			     self->axes[2][2]);
    }
}

-(void) get
{
    const char *k;
    int i, j;
    
    k = lua_tostring (_L, -1);

    if (!xstrcmp(k, "anchor")) {
	dJointGetBallAnchor ([self joint], self->anchor);

        lua_newtable (_L);
        
        for(j = 0; j < 3; j += 1) {
            lua_pushnumber (_L, self->anchor[j]);
            lua_rawseti (_L, -2, j + 1);
        }
    } else if (!xstrcmp(k, "axes")) {
	dJointGetAMotorAxis ([self amotor], 0, self->axes[0]);
	dJointGetAMotorAxis ([self amotor], 1, self->axes[1]);
	dJointGetAMotorAxis ([self amotor], 2, self->axes[2]);

	lua_newtable (_L);
        
	for(j = 0 ; j < 3 ; j += 1) {
	    lua_newtable (_L);
        
	    for(i = 0; i < 3; i += 1) {
		lua_pushnumber (_L, self->axes[j][i]);
		lua_rawseti (_L, -2, i + 1);
	    }

	    lua_rawseti (_L, -2, j + 1);
	}
    } else if (!xstrcmp(k, "motor")) {
	lua_newtable (_L);
        
	for(j = 0 ; j < 3 ; j += 1) {
	    lua_newtable (_L);
        
	    for(i = 0; i < 2; i += 1) {
		lua_pushnumber (_L, self->motor[j][i]);
		lua_rawseti (_L, -2, i + 1);
	    }

	    lua_rawseti (_L, -2, j + 1);
	}
    } else if (!xstrcmp(k, "stops")) {
	lua_newtable (_L);
        
	for(j = 0 ; j < 3 ; j += 1) {
	    lua_newtable (_L);

	    lua_newtable (_L);
	    for(i = 0; i < 2; i += 1) {
		lua_pushnumber (_L, self->stops[j][i]);
		lua_rawseti (_L, -2, i + 1);
	    }
	    lua_rawseti (_L, -2, 1);
	
	    lua_newtable (_L);
	    for(i = 0; i < 2; i += 1) {
		lua_pushnumber (_L, self->hardness[j][i]);
		lua_rawseti (_L, -2, i + 1);
	    }
	    lua_rawseti (_L, -2, 2);

	    lua_pushnumber (_L, bounce[j]);
	    lua_rawseti (_L, -2, 3);

	    lua_rawseti (_L, -2, j + 1);
	}
    } else if (!xstrcmp(k, "tolerance")) {
	lua_newtable (_L);
        
	for(j = 0 ; j < 3 ; j += 1) {
	    lua_pushnumber (_L, self->tolerance[j]);

	    lua_rawseti (_L, -2, j + 1);
	}
    } else if (!xstrcmp(k, "state")) {
	lua_newtable (_L);
        
	for(j = 0 ; j < 3 ; j += 1) {
	    lua_pushnumber (_L, dJointGetAMotorAngle ([self amotor], j));
	    lua_rawseti (_L, -2, j + 1);
	    
	    lua_pushnumber (_L, 0/* dJointGetAMotorAngleRate ([self amotor], j) */);
	    lua_rawseti (_L, -2, j + 4);
	}
    } else {
	[super get];
    }
}

-(void) set
{
    const char *k;
    int i, j;
    
    k = lua_tostring (_L, -2);

    if (!xstrcmp(k, "anchor")) {
        if(lua_istable (_L, 3)) {
            for(j = 0 ; j < 3 ; j += 1) {
                lua_rawgeti (_L, 3, j + 1);
                self->anchor[j] = lua_tonumber (_L, -1);
                
                lua_pop (_L, 1);
            }

	    dJointSetBallAnchor ([self joint],
				 self->anchor[0],
				 self->anchor[1],
				 self->anchor[2]);
        }
    } else if (!xstrcmp(k, "axes")) {
        if(lua_istable (_L, 3)) {
	    dBodyID a, b;
	    
	    for(j = 0 ; j < 3 ; j += 1) {
		lua_rawgeti (_L, 3, j + 1);
		
		if(lua_istable (_L, -1)) {
		    for(i = 0 ; i < 3 ; i += 1) {
			lua_rawgeti (_L, -1, i + 1);
			self->axes[j][i] = lua_tonumber (_L, -1);
                
			lua_pop (_L, 1);
		    }
		}

		dSafeNormalize3 (self->axes[j]);

		lua_pop (_L, 1);
	    }

	    a = dJointGetBody ([self amotor], 0);
	    b = dJointGetBody ([self amotor], 1);

	    dJointSetAMotorAxis ([self amotor], 0, a ? 1 : 0,
				 self->axes[0][0],
				 self->axes[0][1],
				 self->axes[0][2]);

	    dJointSetAMotorAxis ([self amotor], 2, b ? 2 : 0,
				 self->axes[2][0],
				 self->axes[2][1],
				 self->axes[2][2]);
	}
    } else if (!xstrcmp(k, "motor")) {
        if(lua_istable (_L, 3)) {
	    for(j = 0 ; j < 3 ; j += 1) {
		lua_rawgeti (_L, 3, j + 1);
		
		if(lua_istable (_L, -1)) {
		    for(i = 0 ; i < 2 ; i += 1) {
			lua_rawgeti (_L, -1, i + 1);
			self->motor[j][i] = lua_tonumber (_L, -1);
                
			lua_pop (_L, 1);
		    }

		    dJointSetAMotorParam ([self amotor],
					  dParamVel + dParamGroup * j,
					  self->motor[j][0]);
		    
		    dJointSetAMotorParam ([self amotor],
					  dParamFMax + dParamGroup * j,
					  self->motor[j][1]);
		}

		lua_pop (_L, 1);
	    }
	}
    } else if (!xstrcmp(k, "stops")) {
	for(j = 0 ; j < 3 ; j += 1) {
	    /* Resetting the self->stops makes sure that lo remains
	       smaller than hi between calls. */

	    dJointSetAMotorParam ([self amotor],
				  dParamLoStop + dParamGroup * j,
				  -dInfinity);
		    
	    dJointSetAMotorParam ([self amotor],
				  dParamHiStop + dParamGroup * j,
				  dInfinity);
	}
	    
        if(lua_istable (_L, 3)) {
	    for(j = 0 ; j < 3 ; j += 1) {
		lua_rawgeti (_L, 3, j + 1);
		
		if(lua_istable (_L, -1)) {
		    lua_rawgeti (_L, -1, 1);
		    for(i = 0 ; i < 2 ; i += 1) {
			lua_rawgeti (_L, -1, i + 1);
		
			self->stops[j][i] = lua_tonumber (_L, -1);
		
			lua_pop (_L, 1);
		    }
		    lua_pop (_L, 1);

		    lua_rawgeti (_L, -1, 2);
		    for(i = 0 ; i < 2 ; i += 1) {
			lua_rawgeti (_L, -1, i + 1);

			self->hardness[j][i] = lua_tonumber (_L, -1);
			
			lua_pop (_L, 1);
		    }
		    lua_pop (_L, 1);
	
		    lua_rawgeti (_L, -1, 3);
		    bounce[j] = lua_tonumber (_L, -1);
		    lua_pop (_L, 1);
		    
		    dJointSetAMotorParam ([self amotor],
					  dParamLoStop + dParamGroup * j,
					  self->stops[j][0]);

		    dJointSetAMotorParam ([self amotor],
					  dParamHiStop + dParamGroup * j,
					  self->stops[j][1]);

		    dJointSetAMotorParam ([self amotor],
					  dParamStopCFM + dParamGroup * j,
					  self->hardness[j][0]);

		    dJointSetAMotorParam ([self amotor],
					  dParamStopERP + dParamGroup * j,
					  self->hardness[j][1]);

		    dJointSetAMotorParam ([self amotor],
					  dParamBounce + dParamGroup * j,
					  bounce[j]);
		}

		lua_pop (_L, 1);
	    }
	}
    } else if (!xstrcmp(k, "tolerance")) {
        if(lua_istable (_L, 3)) {
            for(j = 0 ; j < 3 ; j += 1) {
                lua_rawgeti (_L, 3, j + 1);
                self->tolerance[j] = lua_tonumber (_L, -1);
                
                lua_pop (_L, 1);

		dJointSetBallParam ([self joint],
				    dParamCFM + dParamGroup * j,
				    self->tolerance[j]);
            }
        }
    } else {
	[super set];
    }
}

-(void) traversePass: (int)pass
{
    if (pass == 2 && self->debug) {
	dBodyID a, b;
	const dReal *q;
	dVector3 p, x, y, z;

	a = dJointGetBody ([self joint], 0);
	b = dJointGetBody ([self joint], 1);

	assert (a || b);

	dJointGetBallAnchor ([self joint], p);
	dJointGetAMotorAxis ([self amotor], 0, x);
	dJointGetAMotorAxis ([self amotor], 1, y);
	dJointGetAMotorAxis ([self amotor], 2, z);

	glUseProgramObjectARB(0);

	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LINE_SMOOTH);
	glEnable(GL_POINT_SMOOTH);
	glEnable(GL_BLEND);
	glDepthMask (GL_FALSE);

	glPushMatrix();
	glTranslatef(p[0], p[1], p[2]);
	
	/* Draw the dots. */

	glPointSize (5);

	glColor3f(1, 0, 0);

	glBegin (GL_POINTS);
	glVertex3f (x[0], x[1], x[2]);
	glEnd ();	

	glColor3f(0, 1, 0);

	glBegin (GL_POINTS);
	glVertex3f (y[0], y[1], y[2]);
	glEnd ();	

	glColor3f(0, 0, 1);

	glBegin (GL_POINTS);
	glVertex3f (z[0], z[1], z[2]);
	glEnd ();	
	
	/* Draw the lines. */

	glLineWidth (1);

	glColor3f(1, 0, 0);

	glBegin (GL_LINES);
	glVertex3f (0, 0, 0);
	glVertex3f (x[0], x[1], x[2]);

	glEnd ();	

	glColor3f(0, 1, 0);

	glBegin (GL_LINES);
	glVertex3f (0, 0, 0);
	glVertex3f (y[0], y[1], y[2]);
	glEnd ();	

	glColor3f(0, 0, 1);

	glBegin (GL_LINES);
	glVertex3f (0, 0, 0);
	glVertex3f (z[0], z[1], z[2]);
	glEnd ();

	glPopMatrix ();

	if (b) {
	    q = dBodyGetPosition (b);

	    /* Draw one arm. */
	    
	    glColor3f(1, 0, 0);
	    glLineWidth (1);

	    glBegin (GL_LINES);
	    glVertex3f (p[0], p[1], p[2]);
	    glVertex3f (q[0], q[1], q[2]);
	    glEnd();
	}

	if (a) {
	    q = dBodyGetPosition (a);

	    /* Draw the other arm and anchor. */
	
	    glColor3f(0, 1, 0);
	    glLineWidth (3);

	    glBegin (GL_LINES);
	    glVertex3f (p[0], p[1], p[2]);
	    glVertex3f (q[0], q[1], q[2]);
	    glEnd();

	    glPointSize (8);
	    
	    glBegin(GL_POINTS);
	    glVertex3f (p[0], p[1], p[2]);
	    glEnd();
	}

	glDepthMask (GL_TRUE);
	glDisable(GL_BLEND);
	glDisable(GL_LINE_SMOOTH);
	glDisable(GL_POINT_SMOOTH);
	glDisable(GL_DEPTH_TEST);
    }
    
    [super traversePass: pass];
}

@end
